/* $Id: passfd.c,v 1.4 1999/11/07 22:56:26 dm Exp $ */

/*
 *
 * Copyright (C) 1998 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */


#include "sysconf.h"

#if !defined (HAVE_ACCRIGHTS) && !defined (HAVE_CMSGHDR)
#error Do not know how to pass file descriptors
#endif /* !HAVE_ACCRIGHTS && !HAVE_CMSGHDR */

#define PASSFD_SELECT 1

static char magicmsg[] = {'\175'};

static void
waitfd (int fd, int write)
{
  static int nfd;
  static fd_set *fds;
  int res;

  assert (fd >= 0);
  if (fd >= nfd) {
    nfd = fd + 1;
    if (nfd < FD_SETSIZE)
      nfd = FD_SETSIZE;
    nfd = (nfd + 0x3f) & ~0x3f;
    if (fds)
      free (fds);
    fds = (fd_set *) malloc (nfd >> 3);
    if (!fds) {
      nfd = 0;
      return;
    }
    bzero (fds, nfd >> 3);
  }

  do {
    FD_SET (fd, fds);
    if (write)
      res = select (fd + 1, NULL, fds, NULL, NULL);
    else
      res = select (fd + 1, fds, NULL, NULL, NULL);
  } while (!res || (res < 0 && errno == EINTR));

  FD_CLR (fd, fds);
}

int
sendfd (int dstfd, int fd)
{
  int res;
  struct msghdr mh;
  struct iovec iov[1];
#ifdef HAVE_CMSGHDR
  char cmhbuf[sizeof (struct cmsghdr) + sizeof (int)];
  struct cmsghdr *const cmh = (struct cmsghdr *) cmhbuf;
  int * const fdp = (int *) (cmhbuf + sizeof (struct cmsghdr));
#else /* !HAVE_CMSGHDR */
  int fdpbuf;
  int *fdp = &fdpbuf;
#endif /* !HAVE_CMSGHDR */

  *fdp = fd;
  bzero (&mh, sizeof mh);
  mh.msg_iov = iov;
  mh.msg_iovlen = 1;
  iov[0].iov_base = magicmsg;
  iov[0].iov_len = sizeof (magicmsg);

#ifdef HAVE_CMSGHDR
  mh.msg_control = (char *) cmh;
  mh.msg_controllen = cmh->cmsg_len = sizeof (cmhbuf);
  cmh->cmsg_level = SOL_SOCKET;
  cmh->cmsg_type = SCM_RIGHTS;
#else /* !HAVE_CMSGHDR */
  mh.msg_accrights = (char *) fdp;
  mh.msg_accrightslen = sizeof (fdpbuf);
#endif /* !HAVE_CMSGHDR */

  waitfd (dstfd, 1);
  res = sendmsg (dstfd, &mh, 0);

  return res < 0 ? -1 : 0;
}

int
recvfd (int srcfd)
{
  struct msghdr mh;
  struct iovec iov[1];
  char msg[sizeof (magicmsg)];
#ifdef HAVE_CMSGHDR
  char cmhbuf[sizeof (struct cmsghdr) + sizeof (int)];
  struct cmsghdr *const cmh = (struct cmsghdr *) cmhbuf;
  int * const fdp = (int *) (cmhbuf + sizeof (struct cmsghdr));
#else /* !HAVE_CMSGHDR */
  int fdpbuf;
  int *fdp = &fdpbuf;
#endif /* !HAVE_CMSGHDR */

  *fdp = -1;
  bzero (&mh, sizeof mh);
  mh.msg_iov = iov;
  mh.msg_iovlen = 1;
  iov[0].iov_base = msg;
  iov[0].iov_len = sizeof (msg);

#ifdef HAVE_CMSGHDR
  mh.msg_control = (char *) cmh;
  mh.msg_controllen = cmh->cmsg_len = sizeof (cmhbuf);
  cmh->cmsg_level = SOL_SOCKET;
  cmh->cmsg_type = SCM_RIGHTS;
#else /* !HAVE_CMSGHDR */
  mh.msg_accrights = (char *) fdp;
  mh.msg_accrightslen = sizeof (fdpbuf);
#endif /* !HAVE_CMSGHDR */

#ifdef PASSFD_SELECT
  waitfd (srcfd, 0);
  errno = 0;
  return recvmsg (srcfd, &mh, 0) == -1 ? -1 : *fdp;
#else /* !PASSFD_SELECT */
  int flags = fcntl (srcfd, F_GETFL);
  if (flags < 0)
    return -1;
  if ((flags & O_NONBLOCK)
      && fcntl (srcfd, F_SETFL, flags & ~O_NONBLOCK) < 0)
    return -1;
  int res = recvmsg (srcfd, &mh, 0) == -1 ? -1 : *fdp;
  if (!(flags & O_NONBLOCK))
    fcntl (srcfd, F_SETFL, flags);
  return res;
#endif /* !PASSFD_SELECT */
}
